# Copyright 2019 Patrick Ulbrich <zulu99@gmx.net>
# Copyright 2019 razer <razerraz@free.fr>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301, USA.
#

from Mailnag.common.dist_cfg import PACKAGE_NAME

try:
	import gi
	gi.require_version('Secret', '1')
	from gi.repository import Secret
	_libsecret_err = None
except ModuleNotFoundError as e:
        _libsecret_err = e



class SecretStore():
	_instance = None
	
	def __init__(self):
		if _libsecret_err != None:
			raise _libsecret_err
	        
		self._schema = Secret.Schema.new(
			f'com.github.pulb.{PACKAGE_NAME}', Secret.SchemaFlags.NONE,
			{'id' : Secret.SchemaAttributeType.STRING})


	def get(self, secret_id):
		return Secret.password_lookup_sync(self._schema, {'id': secret_id}, None)


	def set(self, secret_id, secret, description):
		Secret.password_store_sync(
			self._schema, {'id': secret_id}, Secret.COLLECTION_DEFAULT, 
			description, secret, None)


	def remove(self, secret_id):
		return Secret.password_clear_sync(self._schema, {'id': secret_id}, None)
	
	
	@staticmethod
	def get_default():
		if _libsecret_err != None:
			return None
        	
		if SecretStore._instance == None:
			SecretStore._instance = SecretStore()

		return SecretStore._instance



# Copyright 2013 Patrick Ulbrich <zulu99@gmx.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301, USA.
#

import subprocess
import threading
import logging
import time

# Note : All functions of this module *are* thread-safe.

# Protects access to the proc dictionary.
_proc_lock = threading.Lock()

# Intended to prevent simultaneous execution of 
# (start_subprocess, terminate_subprocesses), 
# (terminate_subprocesses, terminate_subprocesses).
_func_lock = threading.Lock()

# Dictionary holding popen objects 
# and associated thread objects.
_procs = {}


# Starts a subprocess and an associated thread that waits for
# the subprocess to terminate (prevents zombie processes).
def start_subprocess(args, shell = False, callback = None):
	def thread():
		t = threading.currentThread()
		
		try:
			p = None
			with _proc_lock: p = _procs[t]
			retcode = p.wait()
			
			if callback != None:
				callback(retcode)
		finally:
			with _proc_lock: del _procs[t]
	
	with _func_lock:
		p = None
		pid = -1
		try:
			p = subprocess.Popen(args, shell = shell)
		except: logging.exception('Caught an exception.')
	
		if p != None:
			pid = p.pid
			t = threading.Thread(target = thread)
			with _proc_lock:
				_procs[t] = p
				t.start()
	
		return pid


# Terminates all subprocesses that were started by
# start_subprocess(). Subprocesses that don't terminate
# within the timeframe (seconds) specified by the 
# timeout argument, are sent a kill signal.
def terminate_subprocesses(timeout = 3.0):
	with _func_lock:
		
		threads = []
		
		with _proc_lock:
			if len(_procs) == 0:
				return
		
			for t, p in _procs.items():
				threads.append(t)
			
				# Ask all runnig processes to terminate.
				# This will also terminate associated threads 
				# waiting for p.wait().
				# Note : terminate() does not block.
				try: p.terminate()
				except: logging.debug('p.terminate() failed')

		# Start a watchdog thread that will kill 
		# all processes that didn't terminate within <timeout> seconds.
		wd = _Watchdog(timeout)
		wd.start()
	
		# Wait for all threads to terminate
		for t in threads:
			t.join()
	
		wd.stop()
	
		if not wd.triggered:
			logging.info('All subprocesses exited normally.')


# Internal Watchdog class
class _Watchdog(threading.Thread):
	def __init__(self, timeout):
		threading.Thread.__init__(self)
		self.triggered = False
		self._timeout = timeout
		self._event = threading.Event()


	def run(self):
		self._event.wait(self._timeout)
		if not self._event.is_set():
			logging.warning('Process termination took too long - watchdog starts killing...')
			self.triggered = True
			with _proc_lock:
				for t, p in _procs.items():
					try:
						# Kill process p and quit the thread 
						# waiting for p to terminate (p.wait()).
						p.kill()
						logging.info('Watchdog killed process %s' % p.pid)
					except: logging.debug('p.kill() failed')


	def stop(self):
		# Abort watchdog thread (may have been triggered already)
		self._event.set()
		# Wait for watchdog thread (may be inactive already)
		self.join()



# Copyright 2020 Patrick Ulbrich <zulu99@gmx.net>
# Copyright 2016, 2024 Timo Kankare <timo.kankare@iki.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301, USA.
#

"""Implementation of local mailboxes, like mbox and maildir."""

import email
import mailbox
import logging
import os.path

from Mailnag.backends.base import MailboxBackend

class MBoxBackend(MailboxBackend):
	"""Implementation of mbox mail boxes."""
	
	def __init__(self, name = '', path=None, **kw):
		"""Initialize mbox mailbox backend with a name and path."""
		self._name = name
		self._path = path
		self._opened = False


	def open(self):
		"""'Open' mbox. (Actually just checks that mailbox file exists.)"""
		if not os.path.isfile(self._path):
			raise IOError('Mailbox {} does not exist.'.format(self._path))
		self._opened = True


	def close(self):
		"""Close mbox."""
		self._opened = False


	def is_open(self):
		"""Return True if mailbox is opened."""
		return self._opened


	def list_messages(self):
		"""List unread messages from the mailbox.
		Yields pairs (folder, message) where folder is always ''.
		"""
		mbox = mailbox.mbox(self._path, create=False)
		folder = ''
		try:
			for msg in mbox:
				if 'R' not in msg.get_flags():
					yield (folder, msg, {})
		finally:
			mbox.close()


	def request_folders(self):
		"""mbox does not suppoert folders."""
		raise NotImplementedError("mbox does not support folders")


	def supports_mark_as_seen(self):
		return False


	def mark_as_seen(self, mails):
		# TODO: local mailboxes should support this
		raise NotImplementedError

		
	def notify_next_change(self, callback=None, timeout=None):
		raise NotImplementedError("mbox does not support notifications")


	def cancel_notifications(self):
		raise NotImplementedError("mbox does not support notifications")


class MaildirBackend(MailboxBackend):
	"""Implementation of maildir mail boxes."""

	def __init__(self, name = '', path=None, folders=[], **kw):
		"""Initialize maildir mailbox backend with a name, path and folders."""
		self._name = name
		self._path = path
		self._folders = folders
		self._opened = False


	def open(self):
		"""'Open' mailbox. (Actually just checks that maildir directory exists.)"""
		if not os.path.isdir(self._path):
			raise IOError('Mailbox {} does not exist.'.format(self._path))
		self._opened = True


	def close(self):
		"""Close mailbox."""
		self._opened = False


	def is_open(self):
		"""Return True if mailbox is opened."""
		return self._opened


	def list_messages(self):
		"""List unread messages from the mailbox.
		Yields tuples (folder, message, flags).
		"""
		folders = self._folders if len(self._folders) != 0 else ['']
		root_maildir = mailbox.Maildir(self._path, factory=None, create=False)
		try:
			for folder in folders:
				maildir = self._get_folder(root_maildir, folder)
				for msg in maildir:
					if 'S' not in msg.get_flags():
						yield folder, msg, {}
		finally:
			root_maildir.close()


	def request_folders(self):
		"""Lists folders from maildir."""
		maildir = mailbox.Maildir(self._path, factory=None, create=False)
		try:
			return [''] + maildir.list_folders()
		finally:
			maildir.close()


	def mark_as_seen(self, mails):
		# TODO: local mailboxes should support this
		raise NotImplementedError


	def notify_next_change(self, callback=None, timeout=None):
		raise NotImplementedError("maildir does not support notifications")


	def cancel_notifications(self):
		raise NotImplementedError("maildir does not support notifications")


	def _get_folder(self, maildir, folder):
		"""Returns folder instance of the given maildir."""
		if folder == '':
			return maildir
		else:
			return maildir.get_folder(folder)




'''
This awesome piece of code is intented to encode and decode modified UTF-7 
(the one that is used for IMAP folder names)

encode_mutf7(text) - to encode
decode_mutf7(text) - to decode
'''

__author__ = "https://github.com/cheshire-mouse"
__license__ = "WTFPL v. 2"

import base64
import re

ascii_codes = set(range(0x20,0x7f))

def __get_ascii(text):
    pos = 0
    for c in text:
        if ord(c) not in ascii_codes :
            break
        pos += 1
    return text[:pos]

def __remove_ascii(text):
    pos = 0
    for c in text:
        if ord(c) not in ascii_codes :
            break
        pos += 1
    return text[pos:]

def __get_nonascii(text):
    pos = 0
    for c in text:
        if ord(c) in ascii_codes :
            break
        pos += 1
    return text[:pos]

def __remove_nonascii(text):
    pos = 0
    for c in text:
        if ord(c) in ascii_codes :
            break
        pos += 1
    return text[pos:]

def __encode_modified_utf7(text):
    #modified base64 - good old base64 without padding characters (=)
    result = base64.b64encode(text.encode('utf-16be')).decode('utf-8').rstrip('=')
    result = result.replace('/',',')
    result = '&' + result + '-'
    return result

def encode_mutf7(text):
    result = ""
    text = text.replace('&','&-')
    while len(text) > 0:
        result += __get_ascii(text)
        text = __remove_ascii(text)
        if len(text) > 0:
            result += __encode_modified_utf7(__get_nonascii(text))
            text = __remove_nonascii(text)
    return result

def __decode_modified_utf7(text):
    if text == '&-':
        return '&'
    #remove leading & and trailing -
    text_mb64 = text[1:-1]
    text_b64 = text_mb64.replace(',','/')
    #back to normal base64 with padding
    while len(text_b64) % 4 != 0:
        text_b64 += '='
    text_u16 = base64.b64decode(text_b64)
    result = text_u16.decode('utf-16be')
    return result

def decode_mutf7(text):
    rxp = re.compile('&[^&-]*-')
    match = rxp.search(text)
    while ( match ):
        encoded_text = match.group(0)
        decoded_text = __decode_modified_utf7(encoded_text)
        text = rxp.sub(decoded_text,text, count=1)
        match = rxp.search(text)
    result = text
    return result

# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4

